import errno
import json
import os
from pathlib import Path
import polars as pl

from apm4py.event_log import EventLog
from apm4py.utils import get_file_by_pattern


def load_log(
    data_folder: Path, csv_separator: str = ",", use_internal_format=False
) -> EventLog:
    event_path = get_file_by_pattern(data_folder, "*[eE]vent*.csv")
    if event_path is not None:
        event_semantics_path = get_file_by_pattern(data_folder, "*[eE]vent*.json")

        if event_semantics_path is None:
            print(
                "Event CSV import requires an event semantics JSON file in the same directory"
            )
            raise FileNotFoundError(
                errno.ENOENT, os.strerror(errno.ENOENT), "*[eE]vent*.json"
            )

        with open(event_semantics_path, "r") as event_semantics_file:
            event_semantics = json.load(event_semantics_file)

        print(f"Loading {event_path}")
        events_original = pl.scan_csv(event_path, separator=csv_separator)

    else:
        event_path = get_file_by_pattern(data_folder, "*[eE]vent*.parquet")
        if event_path is None:
            print("Couldn't find any event file in the dataset directory")
            raise FileNotFoundError(
                errno.ENOENT,
                os.strerror(errno.ENOENT),
                "*[eE]vent*.csv | *[eE]vent*.parquet",
            )
        print(f"Loading {event_path}")
        events_original = pl.scan_parquet(event_path)

    cases_path = get_file_by_pattern(data_folder, "*[cC]ase*.csv")
    if cases_path is not None:
        case_semantics_path = get_file_by_pattern(data_folder, "*[cC]ase*.json")
        if case_semantics_path is None:
            print(
                "Case CSV import requires an event semantics JSON file in the same directory"
            )
            raise FileNotFoundError(
                errno.ENOENT, os.strerror(errno.ENOENT), "*[cC]ase*.json"
            )

        with open(case_semantics_path, "r") as case_semantics_file:
            case_semantics = json.load(case_semantics_file)

        print(f"Loading {cases_path}")
        cases_original = pl.scan_csv(cases_path, separator=csv_separator)
    else:
        cases_path = get_file_by_pattern(data_folder, "*[cC]ases*.parquet")
        if cases_path is None:
            print("Couldn't find any cases file in the dataset directory")
            raise FileNotFoundError(
                errno.ENOENT,
                os.strerror(errno.ENOENT),
                "*[cC]ases*.csv | *[cC]ases*.parquet",
            )
        print(f"Loading {cases_path}")
        cases_original = pl.scan_parquet(cases_path)

    variants_original = None
    if use_internal_format:
        variants_original = pl.scan_parquet(
            os.path.join(data_folder, "variants.parquet")
        )

    semantics = {"eventSemantics": event_semantics, "caseSemantics": case_semantics}
    (evt_semantics, cse_semantics) = parse_semantics(semantics)

    return EventLog(
        events=events_original,
        event_semantics=evt_semantics,
        cases=cases_original,
        case_semantics=cse_semantics,
        variants=variants_original,
    )


def parse_semantics(semantics: dict[str]):
    def getColumnName(semantics_json, semantics_type, semantic):
        column_names = getColumnNames(semantics_json, semantics_type, semantic)
        column_name = column_names[0] if len(column_names) > 0 else None
        return column_name

    def getTimeFormat(semantics_json, semantics_type, semantic):
        return [
            s["format"]
            for s in semantics_json[semantics_type]
            if s["semantic"] == semantic
        ][0]

    def getColumnNames(semantics_json, semantics_type, semantic):
        return [
            s["name"]
            for s in semantics_json[semantics_type]
            if s["semantic"] == semantic
        ]

    caseid_column = getColumnName(semantics, "eventSemantics", "Case ID")
    action_column = getColumnName(semantics, "eventSemantics", "Action")
    start_column = getColumnName(semantics, "eventSemantics", "Start")
    complete_column = getColumnName(semantics, "eventSemantics", "Complete")
    if complete_column is None:
        complete_column = start_column

    time_format_mapping = {
        "yyyy": "%Y",
        "MM": "%m",
        "dd": "%d",
        "HH": "%H",
        "mm": "%M",
        "ss": "%S",
        ".SSS": "%.3f",
        "Z": "%z",
    }

    time_format = getTimeFormat(semantics, "eventSemantics", "Start")
    for orig_fmt, polars_fmt in time_format_mapping.items():
        time_format = time_format.replace(orig_fmt, polars_fmt)

    categorical_attr = getColumnNames(
        semantics, "eventSemantics", "CategorialAttribute"
    )
    numeric_attr = getColumnNames(semantics, "eventSemantics", "NumericAttribute")

    event_semantics = {
        "caseid_column": caseid_column,
        "action_column": action_column,
        "start_column": start_column,
        "complete_column": complete_column,
        "time_format": time_format,
        "categorical_attr": categorical_attr,
        "numerical_attr": numeric_attr,
    }

    case_semantics = {}
    if "caseSemantics" in semantics:
        cases_caseid_column = getColumnName(semantics, "caseSemantics", "Case ID")
        cases_categorical_attr = getColumnNames(
            semantics, "caseSemantics", "CategorialAttribute"
        )
        cases_numeric_attr = getColumnNames(
            semantics, "caseSemantics", "NumericAttribute"
        )
        case_semantics = {
            "caseid_column": cases_caseid_column,
            "categorical_attr": cases_categorical_attr,
            "numerical_attr": cases_numeric_attr,
        }

    return (event_semantics, case_semantics)



"""
api  for aggregation requests
"""

from typing import Optional

import pandas as pd

from apm4py.aggregation.groupings import create_grouping
from apm4py.aggregation.metrics import create_metric
from apm4py.api import API


def extract_chart_values(jsn):
    """Cast json into data frame and remove artifacts columns.

    No error is raised in case we don't find artifact columns.
    """
    return pd.DataFrame(jsn["chartValues"]).drop(columns=["$type"], errors="ignore")


def normalise_chart_values(df, json_col):
    df_aux = (
        df.explode(json_col)
        .drop(columns=["$type"], errors="ignore")
        .reset_index(drop=True)
    )
    return pd.concat(
        [
            pd.json_normalize(df_aux.loc[:, json_col]).drop(
                columns=["$type"], errors="ignore"
            ),
            df_aux.drop(columns=[json_col]),
        ],
        axis=1,
    )


class AggregationAPI(API):
    def aggregate(
        self,
        log_id: str,
        metric: str,
        grouping: str = None,
        secondary_grouping: str = None,
        max_amount_attributes: int = 10,
        trace_filter_sequence: list = [],
        activity_exclusion_filter: list = [],
        value_sorting: str = "caseCount",
        sorting_order: str = "descending",
        values_from: str = "allCases",
        aggregation_function: Optional[str] = None,
        percentile: Optional[float] = None,
        attribute: Optional[str] = None,
        secondary_attribute: Optional[str] = None,
        activities: Optional[list] = None,
        secondary_activities: Optional[list] = None,
        date_type: Optional[str] = None,
        secondary_date_type: Optional[str] = None,
        **kwargs
    ) -> pd.DataFrame:
        """
        An aggregation function for the computation of KPIs and grouping by
        metrics allowing the creation of bar-charts, line-charts, and other
        visualizations.

        Args:
            log_id:
                A string denoting the id of the log to aggregate.
            metric:
                A string denoting the metric to use. For the value
                "frequency", a frequency metric is returned and for "duration" a
                duration metric is returned. Otherwise the value is interpreted
                as a numeric attribute metric.
            grouping:
                A string denoting the grouping to use. For the value
                "byDuration", a duration grouping is returned. For "byAttribute"
                a grouping by a categorical attribute (the variable attribute
                needs to be passed) is returned. For one of ["byYear", "byMonth",
                "byQuarter", "byDayOfWeek", "byDayOfYear","byHourOfDay"] a time
                grouping is returned (date_type also needs to be set). If the
                activity aggregation "byActivity" is used, the activities to
                aggregate over need to be passed as list.
            secondary_grouping:
                A string denoting the optional secondary grouping to use.
                For the value "byDuration", a duration grouping is returned. For
                "byAttribute" a grouping by a categorical attribute (the variable
                secondary_attribute needs to be passed) is returned. For one of
                ["byYear", "byMonth", "byQuarter", "byDayOfWeek", "byDayOfYear",
                "byHourOfDay"] a time grouping is returned (secpmdary_date_type
                also needs to be set). If the activity aggregation "byActivity"
                is used, the secondary_activities to aggregate over need to be
                passed as list.
            max_amount_attributes:
                An integer denoting the maximum amount of attributes to return.
            trace_filter_sequence:
                A list containing the sequence of filters to apply.
            activity_exclusion_filter:
                A list containing the activities to exclude.
            value_sorting:
                A string denoting the metric to sort the aggregation by.
            sorting_order:
                A string denoting the order of the sorting.
            values_from:
                A string denoting which values to consider for the aggregation.
            aggregation_function:
                An optional string denoting the aggregation function to use for
                numeric attribute metrics.
                Can be one of ["min", "max", "sum", "mean", "median", "variance",
                "standardDeviation"].
            percentile:
                An optional float denoting the percentile to use if instead of
                the available aggregation types listed for aggregation_function
                a percentile aggregation should be used.
            attribute:
                An optional string denoting the attribute to use when grouping
                is set to 'byAttribute'.
            activities:
                An optional list denoting the activities to use when grouping
                is set to 'byActivity'.
            date_type:
                An optional string denoting the date type to use when a time
                grouping is used. It has to be 'startDate' or 'endDate'.
            secondary_attribute:
                An optional string denoting the attribute to use when secondary
                grouping is set to 'byAttribute'.
            secondary_activities:
                An optional list denoting the activities to use when secondary
                grouping is set to 'byActivity'.
            secondary_date_type:
                An optional string denoting the date type to use when a secondary
                time grouping is used. It has to be 'startDate' or 'endDate'.
            **kwargs:
                Keyword arguments passed to requests functions.

        Returns:
            A pandas DataFrame containing the aggregated data.
        """
        request_data = {
            "metric": create_metric(metric, aggregation_function, percentile),
            "valuesFrom": {"type": values_from},
            "miningRequest": {
                "logId": log_id,
                "activityExclusionFilter": activity_exclusion_filter,
                "traceFilterSequence": trace_filter_sequence,
            },
            "options": {
                "maxAmountAttributes": max_amount_attributes,
                "valueSorting": value_sorting,
                "sortingOrder": sorting_order,
            },
        }

        if grouping is not None:
            request_data["grouping"] = create_grouping(
                grouping, date_type, activities, attribute
            )

        if secondary_grouping is not None:
            request_data["secondaryGrouping"] = create_grouping(
                secondary_grouping,
                secondary_date_type,
                secondary_activities,
                secondary_attribute,
            )

        aggregate_response = self.post(
            "/api/v2/aggregate-data", json=request_data, **kwargs
        )

        if aggregate_response.status_code >= 400:
            return pd.DataFrame()

        response_df = extract_chart_values(aggregate_response.json())
        if secondary_grouping is not None:
            response_df = normalise_chart_values(response_df, "values")

        response_df = response_df.rename(
            columns={
                "xAxis": attribute if attribute is not None else grouping,
                "yAxis": metric,
                "zAxis": secondary_attribute
                if secondary_attribute is not None
                else secondary_grouping,
            }
        )
        return response_df

    def boxplot_stats(
        self,
        log_id: str,
        metric: str,
        grouping: str = None,
        attribute: str = None,
        activities: list = None,
        date_type: str = None,
        values_from: str = "allCases",
        **kwargs
    ) -> pd.DataFrame:
        """
        An aggregation function for the computation the metrics necessary for
        building a standard boxplot graph by using the 25th, 50th and 75th
        percentile of the data.

        Args:
            log_id:
                A string denoting the id of the log to aggregate.
            metric:
                A string denoting the metric.
            grouping:
                A string denoting the grouping to use. For the value
                "byDuration", a duration grouping is returned. For "byAttribute"
                a grouping by a categorical attribute (the variable attribute
                needs to be passed) is returned. For one of ["byYear", "byMonth",
                "byQuarter", "byDayOfWeek", "byDayOfYear","byHourOfDay"] a time
                grouping is returned (date_type also needs to be set). If the
                activity aggregation "byActivity" is used, the activities to
                aggregate over need to be passed as list.
            attribute:
                An optional string denoting the attribute to use when grouping
                is set to 'byAttribute'.
            activities:
                An optional list denoting the activities to use when grouping
                is set to 'byActivity'.
            date_type:
                An optional string denoting the date type to use when a time
                grouping is used. It has to be 'startDate' or 'endDate'.
            values_from:
                A string denoting which values to consider for the aggregation.
            **kwargs:
                Keyword arguments passed to aggregate and request function.

        Returns:
            A pandas DataFrame containing the metrics needed for building a boxplot.
        """
        aggregations = [
            self.aggregate(
                log_id=log_id,
                metric=metric,
                grouping=grouping,
                values_from=values_from,
                aggregation_function=function,
                attribute=attribute,
                activities=activities,
                date_type=date_type,
                **kwargs
            )
            for function in ["min", "max", "median"]
        ]
        percentiles = [
            self.aggregate(
                log_id=log_id,
                metric=metric,
                grouping=grouping,
                values_from=values_from,
                percentile=percentile,
                attribute=attribute,
                activities=activities,
                **kwargs
            )
            for percentile in [25, 75]
        ]

        boxplot_stats = pd.DataFrame(
            {
                "min": aggregations[0][metric],
                "max": aggregations[1][metric],
                "median": aggregations[2][metric],
                "p25": percentiles[0][metric],
                "p75": percentiles[1][metric],
            }
        )

        if grouping is not None:
            boxplot_stats.index = aggregations[0][attribute]

        return boxplot_stats



from dataclasses import dataclass
import json
import os
from pathlib import Path
import polars as pl
from polars import col, LazyFrame
import shutil


@dataclass
class EventLog:
    events: LazyFrame
    cases: LazyFrame
    event_semantics: dict[str]
    case_semantics: dict[str]
    variants: LazyFrame = None

    def caseid_events(self):
        return self.event_semantics["caseid_column"]

    def caseid_cases(self):
        return self.case_semantics["caseid_column"]

    def action_col(self):
        return self.event_semantics["action_column"]

    def start_col(self):
        return self.event_semantics["start_column"]

    def save_log(self, path: Path, file_type: str = "csv"):
        if os.path.exists(path):
            shutil.rmtree(path)
        os.makedirs(path)

        match file_type:
            case "parquet":
                self.events.collect().write_parquet(
                    os.path.join(path, "events.parquet")
                )
                self.cases.collect().write_parquet(os.path.join(path, "cases.parquet"))

            case _:
                self.events.collect().write_csv(os.path.join(path, "events.csv"))
                self.cases.collect().write_csv(os.path.join(path, "cases.csv"))

        self.save_semantics(path)

    def save_semantics(self, path: Path):
        with open(
            os.path.join(path, "event_semantics.json"), "w"
        ) as event_semantics_file:
            json.dump(self.event_semantics, event_semantics_file)

        with open(
            os.path.join(path, "case_semantics.json"), "w"
        ) as case_semantics_file:
            json.dump(self.case_semantics, case_semantics_file)

    def compute_variants(self) -> LazyFrame:
        case_id = self.caseid_events()
        action_col = self.action_col()
        start_col = self.start_col()

        variants_raw = (
            self.events.with_columns(
                [
                    col(case_id).cast(pl.Categorical).cast(pl.UInt16),
                    col(action_col).cast(pl.Categorical),
                    col(start_col).str.strptime(
                        pl.Datetime, format=self.event_semantics["time_format"]
                    ),
                ]
            )
            .groupby(case_id)
            .agg([col(action_col).sort_by(start_col).alias("Variant")])
        )

        self.variants = variants_raw.groupby("Variant").agg([pl.count().alias("Freq")])

        return self.variants



"""
Methods for creating the python API for Appian Process Mining.
"""

import warnings
from typing import Optional

from .process_mining_api import ProcessMiningAPI


def create_api(
    scheme: str,
    host: str,
    token: Optional[str] = None,
    instance: int = 1,
    port: int = None,
    application_root: str = None,
    **kwargs,
):
    """Create a configured Lana API.

    The returned api stores the url for a Appian Process Mining
    api as well as your authentication. After creation you can us it to
    manage the Appian process mining resources. Among other things you can
    upload data from python pandas data frames directly or connect logs
    and shiny dashboard resources referencing them by their names.

        api = create_api('https', 'cloud-backend.lanalabs.com', '<a token>')
        upload_response = api.upload_event_log_df(
                                'new-event-log', df_event_log,
                                time_format='YYYY-mm-dd,
                                df_case=df_case_attributes)
        shiny_dashboard = api.create_shiny_dashboard('new-shiny-dashboard')
        connection_response = api.connect_shiny_dashboard(
                                    upload_response.json()['id'],
                                    shiny_dashboard['id'])

    It also provides basic methods to make HTTP verb requests
    directly to the lana endpoints. For example

        response_list = api.get('/api/v2/dashboards')

    Will return a response with a list of dashboard metadata.

    Args:
        scheme:
            A string denoting the scheme of the connection to Lana, usually
            http or https.
        host:
            A string denoting the host of the Lana API.
        token:
            A string representing the user token used for user authentication.
        port:
            (optional) An integer or string denoting the port for the lana
            api. If not set, default ports for the scheme are be used.
        application_root:
            (optional) A string denoting the application root. Only required
            if your lana api is placed outside the URL root, e.g. "/lana-api"
            instead of "/". Has to start with a slash.
        **kwargs:
            Keyword arguments to pass to requests for the initial
            request retrieving user information.
    """

    return ProcessMiningAPI(
        scheme, host, instance, token, port, application_root, **kwargs
    )



"""
functions to prepare requests for consumption
"""

import glob
import os
import pandas as pd
from pathlib import Path
import re
import subprocess
from typing import Iterable, Dict, List, Optional


def _try_to_map(key, value):
    return {key: value} if key is not None else dict()


def create_semantics(
    columns: Iterable[str],
    case_id: str = "id",
    action: Optional[str] = "action",
    start: Optional[str] = "start",
    complete: Optional[str] = "complete",
    numerical_attributes: Iterable[str] = tuple(),
    impact_attributes: Iterable[str] = tuple(),
    descriptive_attributes: Iterable[str] = tuple(),
    time_format: str = "yyyy-MM-dd HH:mm:ss",
) -> List[Dict[str, str]]:
    """
    create semantics including numeric and categorical attributes

    Args:
        columns:
            A list of strings representing the column names of the
            table.
        case_id:
            A string representing the column name for the case id.
        action:
            A string representing the column name for the activities
            (optional).
        start:
            A string representing the column name for the start timestamp
            (optional).
        complete:
            A string representing the column name for the complete timestamp
            (optional).
        numerical_attributes:
            A list of strings representing the column names for numerical
            attributes (optional), defaults to empty tuple.
        impact_attributes:
            A list of strings representing the column names for impact
            attributes (optional), defaults to empty tuple.
        descriptive_attributes:
            A list of strings representing  the column names for
            descriptive attributes (optional), defaults to empty tuple.
        time_format:
            A string representing the time format for start and complete
            columns.

    Returns:
        A list of dictionaries representing the semantics.
    """

    semantics_map = {
        **{case_id: "Case ID"},
        **_try_to_map(action, "Action"),
        **_try_to_map(start, "Start"),
        **_try_to_map(complete, "Complete"),
        **{att: "NumericAttribute" for att in numerical_attributes},
        **{att: "DescriptiveAttribute" for att in descriptive_attributes},
        **{att: "ImpactAttribute" for att in impact_attributes},
    }

    return [
        {
            "idx": idx,
            "name": col,
            "semantic": semantics_map.get(col, "CategorialAttribute"),
            "format": time_format if col in [start, complete] else None,
        }
        for idx, col in enumerate(columns)
    ]


def _is_lana_numeric(ser):
    return pd.api.types.is_numeric_dtype(ser) and not pd.api.types.is_bool_dtype(ser)


def create_event_semantics_from_df(
    df: pd.DataFrame,
    case_id: str = "Case_ID",
    action: str = "Action",
    start: str = "Start",
    complete: str = "Complete",
    impact_attributes: Iterable[str] = tuple(),
    descriptive_attributes: Iterable[str] = tuple(),
    time_format: str = "yyyy-MM-dd'T'HH:mm:ss.SSSSSS",
) -> List[dict]:
    """
    Create event semantics from a pandas data frame.

    The columns semantic for lana will be derived from the data frame's dtypes.
    All types will be converted to categorical attributes, besides numbers,
    impact and descriptive attributes. Descriptive attributes are special
    cases of categorial attributes with the dtype 'object', impact
    attributes are special cases of numerical attributes. Columns with
    either of these semantics can not be inferred. They have to be passed
    to this function.

    The Start and Complete time stamps need to have the same time format.
    For an overview over time stamp formats see
    https://docs.oracle.com/javase/8/docs/api/java/time/format/DateTimeFormatter.html#patterns

    Args:
        df:
            A data frame representing an event log.
        case_id:
            A string representing the column name for the case id.
        action:
            A string representing the column name for the activity.
        start:
            A string representing the column name for the start timestamp.
        complete:
            A string representing the column name for the complete timestamp.
        impact_attributes:
            A list of strings representing the column names for impact
            attributes.
        descriptive_attributes:
            A list of strings representing the column names for descriptive
            attributes.
        time_format:
            A string representing the time format for start and complete
            columns.

    Returns:
        A list of dictionaries representing the log semantics.
    """
    return create_semantics(
        df.columns,
        case_id,
        action,
        start,
        complete,
        numerical_attributes=[
            col
            for col in df.columns
            if _is_lana_numeric(df.loc[:, col])
            and col not in [case_id, action, start, complete]
        ],
        impact_attributes=impact_attributes,
        descriptive_attributes=descriptive_attributes,
        time_format=time_format,
    )


def create_case_semantics_from_df(
    df: pd.DataFrame,
    case_id: str = "Case_ID",
    impact_attributes: Iterable[str] = tuple(),
    descriptive_attributes: Iterable[str] = tuple(),
) -> List[dict]:
    """
    Create event semantics from a pandas data frame.

    The columns semantic for lana will be derived from the data frame's dtypes.
    All types will be converted to categorical attributes, besides numbers,
    impact and descriptive attributes. Descriptive attributes are special
    cases of categorial attributes with the dtype 'object', impact
    attributes are special cases of numerical attributes. Columns with
    either of these semantics can not be inferred. They have to be passed
    to this function.

    Args:
        df:
            A data frame representing amn event log.
        case_id:
            A string representing the column name for the case id.
        impact_attributes:
            A list of strings representing the column names for impact
            attributes.
        descriptive_attributes:
            A list of strings representing the column names for descriptive
            attributes.

    Returns:
        A list of dictionaries representing the log semantics.
    """
    return create_semantics(
        df.columns,
        case_id,
        numerical_attributes=[
            col
            for col in df.columns
            if _is_lana_numeric(df.loc[:, col]) and col != case_id
        ],
        impact_attributes=impact_attributes,
        descriptive_attributes=descriptive_attributes,
    )


def get_semantic(semantics: dict[str], name: str):
    [s["name"] for s in semantics if s["semantic"] == name][0]


def count_lines_in_file(filepath: str) -> int:
    """
    Counts the number of lines in a given file
    See https://gist.github.com/zed/0ac760859e614cd03652 - wccount

    Args:
        filepath:
            Path to the file of which you want to count the number of lines
    """
    out = subprocess.Popen(
        ["wc", "-l", filepath], stdout=subprocess.PIPE, stderr=subprocess.STDOUT
    ).communicate()[0]

    return int(re.search(r"\d+", out.decode()).group())


def get_file_by_pattern(dir: Path, file_pattern: str) -> Optional[str]:
    files = glob.glob(os.path.join(dir, file_pattern))
    if len(files) == 0:
        return None

    if len(files) > 1:
        raise (f"There should be only 1 file matching '{file_pattern}'")

    return files[0]



"""
functions to create the aggregation request grouping
"""

from typing import Optional

from apm4py.aggregation.utils import _raise


def create_by_duration(kind: str) -> dict:
    return {"type": "byDuration"} if kind == "byDuration" else dict()


def create_by_time(kind: str, date_type: str) -> dict:
    intervals = [
        "byYear",
        "byMonth",
        "byQuarter",
        "byDayOfWeek",
        "byDayOfYear",
        "byHourOfDay",
    ]
    return (
        {"type": kind, "dateType": date_type, "timeZone": "Europe/Berlin"}
        if kind in intervals and date_type in ["startDate", "endDate"]
        else dict()
    )


def create_by_activity(kind: str, activities: list) -> dict:
    return (
        {"type": "byActivity", "selectedActivities": activities}
        if kind == "byActivity" and activities
        else dict()
    )


def create_by_attribute(kind: str, attribute: str) -> dict:
    return (
        {"type": kind, "attribute": attribute}
        if kind == "byAttribute" and attribute is not None
        else dict()
    )


def create_by_with_params(
    kind: str, date_type: str, activities: list, attribute: str
) -> dict:
    if [date_type, activities, attribute].count(None) < 2:
        raise Exception(
            "Only one of date_type, activities and attribute can " "be set, not more."
        )
    return (
        create_by_time(kind, date_type)
        or create_by_activity(kind, activities)
        or create_by_attribute(kind, attribute)
    )


def create_grouping(
    kind: str,
    date_type: Optional[str] = None,
    activities: Optional[list] = None,
    attribute: Optional[str] = None,
) -> dict:
    """
    Create a dictionary containing the grouping in a format necessary
    for the aggregation API request.

    Args:
        kind:
            A string denoting the kind of grouping to use. For the value
            "byDuration", a duration grouping is returned. For "byAttribute"
            a grouping by a categorical attribute (the variable attribute
            needs to be passed) is returned. For one of ["byYear", "byMonth",
            "byQuarter", "byDayOfWeek", "byDayOfYear","byHourOfDay"] a time
            grouping is returned (date_type also needs to be set). If the
            activity aggregation "byActivity" is used, the activities to
            aggregate over need to be passed as list.
        date_type:
            An optional string denoting the date type to use when a time
            grouping is used. It has to be 'startDate' or 'endDate'.
        activities:
            An optional list denoting the activities to use when grouping
            kind is set to 'byActivity'.
        attribute:
            An optional string denoting the attribute to use when grouping
            kind is set to 'byAttribute'.

    Returns:
        A dictionary containing the grouping in the right format for the request.
    """
    return (
        create_by_duration(kind)
        or create_by_with_params(kind, date_type, activities, attribute)
        or _raise(
            Exception(
                f"Impossible to create grouping for {kind}. Make "
                f"sure that a valid aggregation type was used and "
                f"that an attribute was supplied for attribute groupings, "
                f"a date type was supplied for time groupings or "
                f"activities were supplied for activity groupings."
            )
        )
    )



"""
log management api requests functions and methods
"""

import io
import json
import math
from pathlib import Path
from typing import Union, List, TextIO, BinaryIO, Optional, Iterable

import pandas as pd
from requests import Response
from tqdm import tqdm

from apm4py.decorators import expect_json
from apm4py.decorators import handle_response
from apm4py.resources import ResourceAPI
from apm4py.utils import (
    create_case_semantics_from_df,
    create_event_semantics_from_df,
    count_lines_in_file,
)


def _serialise_semantics(semantics: Union[str, list]):
    return json.dumps(semantics) if not isinstance(semantics, str) else semantics


class LogsAPI(ResourceAPI):
    def list_logs(self, **kwargs) -> list:
        """List all logs that are available to the user.

        Args:
            **kwargs:
                Keyword arguments passed to requests functions.

        Returns:
            A list of log names.
        """
        return self.list_resources("logs", **kwargs)

    @expect_json
    def list_user_logs(self, **kwargs) -> list:
        """List all logs owned by the user.

        Args:
            **kwargs:
                Keyword arguments passed to requests functions.

        Returns:
            A list of log names.
        """
        return self.get("/api/users/" + self.user.user_id + "/logs", **kwargs)

    def get_log_ids(self, contains: str = ".*", **kwargs) -> List[str]:
        """Get all log ids which names are matched by the passed regular
        expression.

        Args:
            contains:
                A string denoting a regular expression
                matched against the log names.
            **kwargs:
                Keyword arguments passed to requests functions.

        Returns:
            A list of strings denoting log ids.
        """
        return self.get_resource_ids("logs", contains, **kwargs)

    def get_log_id(self, contains: str, **kwargs) -> str:
        """Get id of a log by its name.

        The name needs to be unique, otherwise an exception is raised.

        Args:
            contains:
                A string denoting a regular expression
                matched against the log names.
            **kwargs:
                Keyword arguments passed to requests functions.

        Returns:
            A string denoting the id of the log.
        """
        return self.get_resource_id("logs", contains, **kwargs)

    def describe_log(self, contains: str = None, log_id: str = None, **kwargs) -> dict:
        """Get the description of log.

        Args:
            log_id:
                A string denoting the id of the log, takes precedence
                over contains.
            contains:
                A string denoting a regular expression matched against
                the log names, matching several names raises an exception.
            **kwargs:
                Keyword arguments passed to requests functions.

        Returns:
            A dictionary denoting the log description.
        """
        return self.describe_resource("logs", contains, log_id, **kwargs)

    def upload_event_log(
        self,
        name: str,
        log: str,
        log_semantics: Union[str, List[dict]],
        case_attributes: Optional[str] = None,
        case_attribute_semantics: Optional[Union[str, List[dict]]] = None,
        logId: str = "",
        **kwargs,
    ) -> Response:
        """Upload an event log with prepared semantics.

        Args:
            name:
                A string denoting the name for the uploaded log.
            log:
                A string denoting the event log as csv.
            log_semantics:
                The event log semantics either serialised as a
                string or a list of dictionaries.
            case_attributes:
                (optional) A string denoting the case
                attributes as csv.
            case_attribute_semantics:
                (optional) The event case attributes semantics
                either serialised as a string or a list of
                dictionaries, required if case_attributes is
                passed.
            logId: (optional) ID of the log which content is to be replaced
            **kwargs:
                Keyword arguments passed to requests functions.

        Returns:
            The requests response of the lana api call.
        """

        file_attr_name = "eventCSVFile" if case_attributes else "file"
        files_required = {file_attr_name: (name, log, "text/csv")}
        semantics_required = {
            "eventSemantics": _serialise_semantics(log_semantics),
            "logName": name,
            "timeZone": "Europe/Berlin",
        }

        files = (
            {
                **files_required,
                **{
                    "caseAttributeFile": (
                        name + "_case_attributes",
                        case_attributes,
                        "text/csv",
                    )
                },
            }
            if case_attributes
            else files_required
        )
        semantics = (
            {
                **semantics_required,
                **{"caseSemantics": _serialise_semantics(case_attribute_semantics)},
            }
            if case_attribute_semantics
            else semantics_required
        )

        if logId == "" and case_attributes:
            endpoint = "csv-case-attributes-event-semantics"
        elif logId == "":
            endpoint = "csv"
        else:
            endpoint = logId + "/replace-log"

        return self.post("/api/logs/" + endpoint, files=files, data=semantics, **kwargs)

    def upload_event_log_stream(
        self,
        log: Union[TextIO, BinaryIO],
        log_semantics: Union[list, str],
        case: Optional[Union[TextIO, BinaryIO]] = None,
        case_semantics: Optional[Union[list, str]] = None,
        prefix: str = "apm4py-",
        **kwargs,
    ) -> Response:
        """Upload a log with prepared semantics by passing open streams.

        The log name is generated from hash value of the passed event log
        stream. We use the built-in hash function, so it can change when you
        restart the interpreter.

        WARNING: This method does not close the passed streams.

        Args:
            log:
                A string or binary stream denoting the event log.
            log_semantics:
                The event log semantics either serialised as a
                string or a list of dictionaries.
            case:
                (optional) A text or binary stream denoting the
                case attributes as csv.
            case_semantics:
                (optional) The event case attributes semantics
                either serialised as a string or a list of
                dictionaries, required if case_attributes is
                passed.
            prefix:
                (optional) A string denoting a prefix of the log name.
                Defaults to "apm4py-".
            **kwargs:
                Keyword arguments passed to requests functions.

        Returns:
            The requests response of the lana api call.
        """

        name = f"{prefix}{hash(log)}"
        return self.upload_event_log(
            name, log.read(), log_semantics, case.read(), case_semantics, **kwargs
        )

    def upload_event_log_df(
        self,
        name: str,
        df_log: pd.DataFrame,
        time_format: str,
        df_case: pd.DataFrame,
        impact_attributes: Iterable[str] = tuple(),
        descriptive_attributes: Iterable[str] = tuple(),
        **kwargs,
    ) -> Response:
        """Upload an event log from pandas data frames with inferred semantics.

        For the passed event log data frame we expect at least the following
        columns:
        - "Case_ID" for the case id, any dtype
        - "Action" of dtype object for activities
        - "Start" of dtype datetime64 or object for the first timestamp
        The "Complete" column of type datetime64 or object is optional.

        For the passed case attributes we expect at last the columns "Case_ID"
        of any dtype.

        The columns semantic for lana will be derived from the data frame's dtypes.
        All types will be converted to categorical attributes, besides numbers,
        impact and descriptive attributes. Descriptive attributes are special
        cases of categorial attributes with the dtype 'object', impact
        attributes are special cases of numerical attributes. Columns with
        either of these semantics can not be inferred. They have to be passed
        to this function.


        Args:
            name:
                A string denoting the name under which the log is uploaded.
            df_log:
                A pandas data frame denoting the event log.
            time_format:
                A string denoting the timestamp format. Needs to match the
                string format of the "Start" and, if present, "Complete"
                columns. If either or both columns are of dtype datetime64,
                the time format needs to match its string representation.
                For a format definition, see
                https://docs.oracle.com/javase/8/docs/api/java/time/format/DateTimeFormatter.html#patterns
            df_case:
                (Optional) A pandas data frame denoting the case
                attributes.
            impact_attributes:
                (optional) A list of strings representing the column names
                for impact attributes, defaults to empty tuple.
            descriptive_attributes:
                (optional) A list of strings representing  the column names for
                descriptive attributes, defaults to empty tuple.
            **kwargs:
                Keyword arguments passed to requests functions.

            Returns:
                The requests response of the lana api call.
        """

        log_semantics = create_event_semantics_from_df(
            df_log,
            time_format=time_format,
            impact_attributes=impact_attributes,
            descriptive_attributes=descriptive_attributes,
        )
        case_semantics = create_case_semantics_from_df(
            df_case,
            impact_attributes=impact_attributes,
            descriptive_attributes=descriptive_attributes,
        )

        return self.upload_event_log(
            name,
            log=df_log.to_csv(index=False),
            log_semantics=log_semantics,
            case_attributes=df_case.to_csv(index=False),
            case_attribute_semantics=case_semantics,
            **kwargs,
        )

    def upload_event_log_file(
        self,
        name: str,
        event_file_path: str,
        case_file_path: str,
        event_semantics_path: str,
        case_semantics_path: str,
        chunk_size: int = 250000,
        log_id: str = "",
        show_progress: bool = False,
        **kwargs,
    ) -> Response:
        """Upload an event log with case attributes by their path.

        Paths to semantic files have to be passed as well. All files are read
        in as binaries, the lana backend will try to infer the encoding.

        Args:
            name:
                A string denoting the name under which the
                log will be uploaded.
            event_file_path:
                A string denoting the path to the event log csv file.
            case_file_path:
                A string denoting the path to the case attribute csv file.
            event_semantics_path:
                A string denoting the path to the event log semantics
                json file.
            case_semantics_path:
                A string denoting the path to the case attributes semantics
                json file.
            chunk_size:
                (optional) Number of lines to upload in a single chunk.
                Set to -1 to upload without chunking.
            **kwargs:
                Keyword arguments passed to requests functions.

        Returns:
            The requests response of the lana api call.
        """

        num_events = count_lines_in_file(event_file_path) if chunk_size > 0 else 0

        if chunk_size > 0 and chunk_size < num_events:
            num_chunks = math.ceil(num_events / chunk_size)
            num_cases = count_lines_in_file(case_file_path)
            print(
                f"Uploading {num_events} events and {num_cases} cases in {num_chunks} chunks"
            )
            if show_progress:
                progress = tqdm(total=num_chunks)

            line_number = 1
            chunk = 1
            event_buffer = []
            case_buffer = []

            with (
                open(event_file_path, "r") as event_file,
                open(event_semantics_path, "r") as event_semantics_file,
                open(case_file_path, "r") as case_file,
                open(case_semantics_path, "r") as case_semantics_file,
            ):
                event_semantics = event_semantics_file.read()
                case_semantics = case_semantics_file.read()
                for line in event_file:
                    event_buffer.append(line)
                    if line_number <= num_cases:
                        case_line = case_file.readline()
                        case_buffer.append(case_line)

                        if line_number == 1:
                            event_header = line
                            case_header = case_line
                            line_number += 1
                            continue

                    if line_number % chunk_size == 0 or line_number == num_events:
                        event_log_chunk = "".join(event_buffer)
                        if line_number == chunk_size:
                            # Initial upload of first chunk
                            log_id = self.upload_event_log(
                                name=name,
                                log=event_log_chunk,
                                log_semantics=event_semantics,
                                case_attributes="".join(case_buffer),
                                case_attribute_semantics=case_semantics,
                                logId=log_id,
                            ).json()["logId"]
                        else:
                            # Incremental upload
                            file = {
                                "eventCSVFile": (
                                    "event-file",
                                    event_log_chunk,
                                    "text/csv",
                                )
                            }
                            semantics = {"eventSemantics": event_semantics}

                            result = self.post(
                                "/api/logs/" + log_id + "/csv",
                                files=file,
                                data=semantics,
                                **kwargs,
                            )

                            if len(case_buffer) > 1:
                                file = {
                                    "caseAttributeFile": (
                                        "case-file",
                                        "".join(case_buffer),
                                        "text/csv",
                                    )
                                }
                                semantics = {"caseSemantics": case_semantics}

                                result = self.post(
                                    "/api/logs/" + log_id + "/csv-case-attributes",
                                    files=file,
                                    data=semantics,
                                )
                        if show_progress:
                            progress.update()
                        chunk += 1
                        event_buffer = [event_header]
                        case_buffer = [case_header]

                    line_number += 1

            if show_progress:
                progress.close()

            # with (
            #     open(case_file_path, "r") as case_file,
            #     open(case_semantics_path, "r") as case_semantics_file,
            # ):
            #     case_semantics = case_semantics_file.read()
            #     for line in case_file:
            #         case_buffer.append(line.replace(";", ","))
            #         i += 1
            #         if i % chunk_size == 0 or i == num_cases:
            #             case_log_chunk = "".join(case_buffer)
            #             if i == chunk_size:
            #                 # Initial upload of first chunk
            #                 log_id = self.upload_case_log(
            #                     name=name,
            #                     log=case_log_chunk,
            #                     log_semantics=case_semantics,
            #                     logId=log_id,
            #                 ).json()["logId"]
            #                 print("Uploaded initial chunk. LogId = " + log_id)
            #             else:
            #                 print(f"Uploaded chunk {chunk}/{num_chunks}")
            #                 # Incremental upload
            #                 file = {
            #                     "caseCSVFile": (
            #                         "case-file",
            #                         case_log_chunk,
            #                         "text/csv",
            #                     )
            #                 }
            #                 semantics = {"caseSemantics": case_semantics}

            #                 result = self.post(
            #                     "/api/logs/" + log_id + "/csv",
            #                     files=file,
            #                     data=semantics,
            #                     **kwargs,
            #                 )

            #             chunk += 1

            #     return result

        else:
            with (
                open(event_file_path, "rb") as event_file,
                open(case_file_path, "rb") as case_file,
                open(event_semantics_path) as event_semantics,
                open(case_semantics_path) as case_semantics,
            ):
                files = {
                    "eventCSVFile": (
                        Path(event_file_path).name,
                        event_file,
                        "text/csv",
                    ),
                    "caseAttributeFile": (
                        Path(case_file_path).name,
                        case_file,
                        "text/csv",
                    ),
                }

                semantics = {
                    "eventSemantics": event_semantics.read(),
                    "caseSemantics": case_semantics.read(),
                    "logName": name,
                    "timeZone": "Europe/Berlin",
                }

                result = self.post(
                    "/api/logs/csv-case-attributes-event-semantics",
                    files=files,
                    data=semantics,
                    **kwargs,
                )

        return result

    def append_events_df(
        self,
        log_id,
        df_log: pd.DataFrame,
        time_format: str,
        impact_attributes: Iterable[str] = tuple(),
        descriptive_attributes: Iterable[str] = tuple(),
        **kwargs,
    ) -> Response:
        """Append events to a log from a pandas data frame.

        For the passed event log data frame we expect at least the following
        columns:
        * "Case_ID" for the case id, any dtype
        * "Action" of dtype object for activities
        * "Start" of dtype datetime64 or object for the first timestamp
        The "Complete" column of type datetime64 or object is optional.

        The columns semantic for lana will be derived from the data frame's dtypes.
        All types will be converted to categorical attributes, besides numbers,
        impact and descriptive attributes. Descriptive attributes are special
        cases of categorial attributes with the dtype 'object', impact
        attributes are special cases of numerical attributes. Columns with
        either of these semantics can not be inferred. They have to be passed
        to this function.

        Args:
            log_id:
                A string denoting the id of the log to which the events
                should be appended.
            df_log:
                A pandas data frame denoting the event log.
            time_format:
                A string denoting the timestamp format. Needs to match the
                string format of the "Start" and, if present, "Complete"
                columns. If either or both columns are of dtype datetime64,
                the time format needs to match its string representation.
                For a format definition, see
                https://docs.oracle.com/javase/8/docs/api/java/time/format/DateTimeFormatter.html#patterns
            impact_attributes:
                (optional) A list of strings representing the column names
                for impact attributes, defaults to empty tuple.
            descriptive_attributes:
                (optional) A list of strings representing  the column names for
                descriptive attributes, defaults to empty tuple.
            **kwargs:
                Keyword arguments passed to requests functions.

            Returns:
                The requests response of the lana api call.
        """
        event_semantics = create_event_semantics_from_df(
            df_log,
            time_format=time_format,
            impact_attributes=impact_attributes,
            descriptive_attributes=descriptive_attributes,
        )

        files = {"eventCSVFile": ("event-file", df_log.to_csv(index=False), "text/csv")}
        semantics = {"eventSemantics": _serialise_semantics(event_semantics)}

        return self.post(
            "/api/logs/" + log_id + "/csv", files=files, data=semantics, **kwargs
        )

    def append_case_attributes_df(
        self, log_id, df_case: pd.DataFrame, **kwargs
    ) -> Response:
        """Append case attributes to a log from a pandas data frame.

        For the passed case attributes we expect at least the case id
        column named "Case_ID".

        Args:
            log_id:
                A string denoting the id of the log for which the case
                attributes should be appended.
            df_case:
                A pandas data frame denoting the case attributes.
            **kwargs:
                Keyword arguments passed to requests functions.

        Returns:
            The requests response of the lana api call.
        """
        case_semantics = create_case_semantics_from_df(df_case)

        files = {
            "caseAttributeFile": ("event-file", df_case.to_csv(index=False), "text/csv")
        }
        semantics = {"caseSemantics": _serialise_semantics(case_semantics)}

        return self.post(
            "/api/logs/" + log_id + "/csv-case-attributes",
            files=files,
            data=semantics,
            **kwargs,
        )

    def replace_log(
        self,
        name_log: str,
        df_log: pd.DataFrame,
        time_format: str,
        df_attr: pd.DataFrame = None,
        impact_attributes: Iterable[str] = tuple(),
        chunk_size: int = 3000000,
        upload_chunk_delay=0,
    ):
        try:
            logId = self.get_log_id(contains=make_strict_re(name_log))
        except Exception:
            logId = ""

        return self.upload_event_log_df(
            self,
            name_log,
            df_log,
            time_format,
            df_attr,
            impact_attributes=impact_attributes,
            logId=logId,
            chunk_size=chunk_size,
            upload_chunk_delay=upload_chunk_delay,
        )

    def delete_log(self, log_id: str, **kwargs) -> Response:
        """Delete a log by its id.

        Args:
            log_id:
                A string denoting the id of the log.
            **kwargs:
                Keyword arguments passed to requests functions.

        Returns:
            The requests response of the lana api call.
        """
        return self.delete_resource("logs", log_id, **kwargs)

    def delete_logs(
        self, log_ids: List[str] = None, contains: str = None, **kwargs
    ) -> List[Response]:
        """Delete one or multiple logs.

        Args:
            log_ids:
                A list of strings denoting the ids of logs to delete. Tales
                precedence over contains.
            contains:
                A string denoting a regular expression matched against
                the log names.
            **kwargs:
                Keyword arguments passed to requests functions.

        Returns:
            The requests response of the lana api call.
        """
        return self.delete_resources("logs", contains, log_ids, **kwargs)

    def request_event_csv(
        self, log_id: str, mining_request: Optional[dict] = None, **kwargs
    ) -> Response:
        """Request the enriched event log.

        Args:
            log_id:
                A string denoting the id of the log.
            mining_request:
                (optional) A mining request data structure sent with
                request to filter down the log.
            **kwargs:
                Keyword arguments passed to requests functions.

        Returns:
            The requests response of the lana api call. The event log can be
            accessed under the text attribute of the response.
        """
        request_field = (
            json.dumps(mining_request)
            if mining_request
            else json.dumps(
                {
                    "activityExclusionFilter": [],
                    "includeHeader": True,
                    "includeLogId": False,
                    "logId": log_id,
                    "edgeThreshold": 1,
                    "traceFilterSequence": [],
                    "runConformance": True,
                    "graphControl": {
                        "sizeControl": "Frequency",
                        "colorControl": "AverageDuration",
                    },
                }
            )
        )
        return self.get(f"/api/eventCsvWithFilter?request={request_field}", **kwargs)

    def get_event_log(
        self,
        log_name: str = None,
        log_id: str = None,
        mining_request: Optional[dict] = None,
        **kwargs,
    ) -> pd.DataFrame:
        """Get the enriched event log as a pandas data frame

        Only columns with time stamps are type cast, the other columns
        remain objects. If type casting is not desired, use to the method
        'request_event_csv'.

        Args:
            log_id:
                A string denoting the id of the log. Tales precedence
                over log_name.
            log_name:
                A string denoting the name of the log.
            mining_request:
                (optional) A mining request data structure sent with
                request to filter down the log.
            **kwargs:
                Keyword arguments passed to requests functions.

        Returns:
            A data frame denoting the enriched log.
        """

        log_id = log_id or self.get_log_id(log_name)
        resp = self.request_event_csv(log_id, mining_request, **kwargs)
        if resp.status_code >= 400:
            return pd.DataFrame()
        csv_stream = io.BytesIO(resp.content)
        return pd.read_csv(csv_stream, dtype="object")

    @handle_response
    def share_log(self, log_id: str) -> Response:
        """Share log with organisation.

        Args:
            log_id:
                A string denoting the log id.

        Returns:
            The requests response of the lana api call.
        """
        return self.get(f"/api/shareLogWithOrg/{log_id}")

    @handle_response
    def unshare_log(self, log_id: str) -> Response:
        """Un-share log with organisation.

        Args:
            log_id:
                A string denoting the log id.

        Returns:
            The requests response of the lana api call.
        """
        return self.get(f"/api/unshareLogWithOrg/{log_id}")

    # legacy methods
    # --------------

    def uploadEventLog(self, logFile, logSemantics):
        file = {
            "file": open(logFile, "rb"),
        }

        semantics = {
            "eventSemantics": open(logSemantics).read(),
        }

        return self.post("/api/logs/csv", files=file, data=semantics)

    def uploadEventLogWithCaseAttributes(
        self,
        logFile,
        logSemantics,
        caseAttributeFile,
        caseAttributeSemantics,
        logName=None,
    ):
        files = {
            "eventCSVFile": (Path(logFile).name, open(logFile, "rb"), "text/csv"),
            "caseAttributesFile": (
                Path(caseAttributeFile).name,
                open(caseAttributeFile, "rb"),
                "text/csv",
            ),
        }

        semantics = {
            "eventSemantics": open(logSemantics).read(),
            "caseSemantics": open(caseAttributeSemantics).read(),
            "logName": logName,
            "timeZone": "Europe/Berlin",
        }

        return self.post(
            "/api/logs/csv-case-attributes-event-semantics", files=files, data=semantics
        )

    def getUserLogs(self):
        return self.list_user_logs()

    def chooseLog(self, logName):
        userLogs = self.getUserLogs()
        logId = max([x["id"] for x in userLogs if x["name"] == logName])
        return logId

    @handle_response
    def appendEvents(self, logId: str, logFile, logSemantics):
        file = {"eventCSVFile": open(logFile, "rb")}
        semantics = {"eventSemantics": open(logSemantics).read()}

        return self.post("/api/logs/" + logId + "/csv", files=file, data=semantics)

    @handle_response
    def appendAttributes(self, logId, caseAttributeFile, caseAttributeSemantics):
        file = {"caseAttributeFile": open(caseAttributeFile, "rb")}
        semantics = {"caseSemantics": open(caseAttributeSemantics).read()}

        return self.post(
            "/api/logs/" + logId + "/csv-case-attributes", files=file, data=semantics
        )



#!/usr/bin/env python

import polars as pl
from polars import col
import os
from os import path
import shutil
import typer
from typing import List, Optional


app = typer.Typer()
pl.toggle_string_cache(True)


DATASETS = {
    "P2P Benchmark": [
        {"name": "medium", "num_events": 8350208, "num_partitions": 2},
        {"name": "large", "num_events": 33400832, "num_partitions": 8},
        {"name": "2xlarge", "num_events": 66801664, "num_partitions": 16},
        {"name": "4xlarge", "num_events": 133603328, "num_partitions": 32},
        {"name": "8xlarge", "num_events": 267206656, "num_partitions": 64},
        {"name": "16xlarge", "num_events": 534413312, "num_partitions": 128},
        {"name": "32xlarge", "num_events": 1068826624, "num_partitions": 256},
    ],
    "P2P Records": [
        {"name": "500k_events", "num_events": 16309 * 31, "num_partitions": 1},
        {"name": "1mio_events", "num_events": 16309 * 62, "num_partitions": 1},
        {"name": "2mio_events", "num_events": 16309 * 123, "num_partitions": 1},
        {"name": "5mio_events", "num_events": 16309 * 62 * 5, "num_partitions": 1},
    ],
}


@app.command()
def grow_log(
    events_original: pl.DataFrame,
    events_case_id: str,
    cases_original: pl.DataFrame,
    cases_case_id: str,
    variants_original: pl.LazyFrame,
    dataset_collection: str,
    output_csv: bool,
    output_feather: bool,
    feather_compression: str,
    compression_level: int,
    data_folder: str,
    dataset: Optional[List[str]],
    partitions: int,
    use_internal_format: bool,
    growth_factor: int,
) -> str:
    if use_internal_format:
        events_original = events_original.with_columns(
            [
                col("CaseID").cast(pl.UInt32),
                col("NextCaseID").cast(pl.UInt32),
            ]
        )
        cases_original = cases_original.with_columns([col("CaseID").cast(pl.UInt32)])

    num_cases = events_original.select(events_case_id).unique().height
    num_events = events_original.height
    print(f"Original dataset has {num_events} events and {num_cases} cases.")

    if dataset_collection is not None:
        datasets = DATASETS[dataset_collection]
        if dataset:
            selected_datasets = []
            for d in dataset:
                selected_datasets.append(next(x for x in datasets if x["name"] == d))
        else:
            selected_datasets = datasets
    else:
        growth_factor = growth_factor if growth_factor is not None else 2
        selected_datasets = [
            {
                "name": dataset[0] if len(dataset) > 0 else f"{growth_factor}xdataset",
                "num_events": growth_factor * num_events,
                "num_partitions": partitions if partitions is not None else 1,
            }
        ]

    file_extension = "parquet"
    if output_csv:
        file_extension = "csv"
    elif output_feather:
        file_extension = "feather"
        if feather_compression != "uncompressed":
            file_extension += f".{feather_compression}"

    for dataset in selected_datasets:
        num_partitions = (
            partitions if partitions is not None else dataset["num_partitions"]
        )
        dataset_path = f"{data_folder}/{dataset['name']}"
        if os.path.exists(dataset_path):
            shutil.rmtree(dataset_path)
        os.makedirs(dataset_path)
        if num_partitions > 1:
            os.makedirs(f"{dataset_path}/events.{file_extension}")
            os.makedirs(f"{dataset_path}/cases.{file_extension}")

        growth_factor = int(dataset["num_events"] / num_events)
        partition_size = int(growth_factor / num_partitions)
        partition_index = 1
        print(
            f"Create dataset {dataset['name']} with {dataset['num_events']} (growth factor = {growth_factor}, partition size = {partition_size}*{num_events}={partition_size*num_events})"
        )
        event_dfs = [events_original]
        cases_dfs = [cases_original]
        for i in range(growth_factor - 1):
            print(f"{int((i+1)/(growth_factor-1)*100)} %", end="\r", flush=True)
            if use_internal_format:
                case_id_add = num_cases * (i + 1) + 1
                case_id_new = col(events_case_id) + case_id_add
                internal_cols = (
                    [col("Action"), col("Start"), (col("NextCaseID") + case_id_add)],
                )
                internal_col_names = ["Action", "Start", "NextCaseID"]
            else:
                case_id_new = pl.concat_str([col(events_case_id), pl.lit(f"_{i}")])
                internal_cols = []
                internal_col_names = []

            event_dfs.append(
                events_original.select(
                    [
                        case_id_new,
                        *internal_cols,
                        pl.all().exclude([events_case_id, *internal_col_names]),
                    ]
                )
            )

            if use_internal_format:
                case_id_add = num_cases * (i + 1) + 1
                case_id_new = col(cases_case_id) + case_id_add
            else:
                case_id_new = pl.concat_str([col(cases_case_id), pl.lit(f"_{i}")])

            cases_dfs.append(
                cases_original.select(
                    [
                        case_id_new,
                        pl.all().exclude(cases_case_id),
                    ]
                )
            )
            if (i + 1) % partition_size == 0 or (i + 1) == (growth_factor - 1):
                if num_partitions > 1:
                    events_partition_path = f"{dataset_path}/events.{file_extension}/events_{partition_index}.{file_extension}"
                    cases_partition_path = f"{dataset_path}/cases.{file_extension}/cases_{partition_index}.{file_extension}"
                else:
                    events_partition_path = f"{dataset_path}/events.{file_extension}"
                    cases_partition_path = f"{dataset_path}/cases.{file_extension}"

                events_grown = pl.concat(event_dfs)
                print(f"Write {events_grown.height} rows to {events_partition_path}")
                if not (output_csv or output_feather):
                    events_grown.write_parquet(
                        events_partition_path, compression_level=compression_level
                    )
                elif output_csv:
                    if not use_internal_format:
                        events_export = events_grown
                    else:
                        events_export = events_grown.drop(
                            ["Duration", "NextCaseID", "NextStart"]
                        )
                    events_export.write_csv(events_partition_path)
                else:
                    events_grown.write_ipc(
                        events_partition_path, compression=feather_compression
                    )

                cases_grown = pl.concat(cases_dfs)
                print(f"Write {cases_grown.height} rows to {cases_partition_path}")

                if not (output_csv or output_feather):
                    cases_grown.write_parquet(
                        cases_partition_path, compression_level=compression_level
                    )
                elif output_csv:
                    cases_grown.write_csv(cases_partition_path)
                else:
                    cases_grown.write_ipc(
                        cases_partition_path, compression=feather_compression
                    )
                partition_index += 1
                event_dfs = []
                cases_dfs = []
        if use_internal_format and not (output_csv or output_feather):
            variants_original.collect().write_parquet(
                f"{dataset_path}/variants.{file_extension}"
            )
        elif use_internal_format and output_feather:
            variants_original.collect().write_ipc(
                f"{dataset_path}/variants.{file_extension}",
                compression=feather_compression,
            )

    return dataset_path


@app.command()
def grow_ocel_csv(
    dataset: str = typer.Argument(help="Name of the dataset collection"),
    data_folder: str = typer.Option("./data", help="Data folder"),
):
    dataset_folder = path.join(data_folder, dataset)
    events_original = pl.scan_csv(path.join(dataset_folder, "events.csv"))


if __name__ == "__main__":
    app()



"""
functions to manipulate event logs to create more complex process data
"""

from random import randrange
import polars as pl


def remove_events(events: pl.LazyFrame, number_events_remove: int) -> pl.LazyFrame:
    number_events = events.select(pl.count()).collect()[0, 0]
    new_events = events.collect()
    for _ in range(number_events_remove):
        row_num = randrange(number_events)
        new_events = new_events.slice(0, row_num - 1).vstack(
            new_events.slice(row_num, None)
        )

    return new_events.lazy()


def duplicate_events(
    events: pl.LazyFrame, number_events_duplicate: int
) -> pl.LazyFrame:
    number_events = events.select(pl.count()).collect()[0, 0]
    new_events = events.collect()
    for _ in range(number_events_duplicate):
        row_num = randrange(number_events)
        new_events = (
            new_events.slice(0, row_num)
            .vstack(new_events.slice(row_num, 1))
            .vstack(new_events.slice(row_num, None))
        )

    return new_events.lazy()


def swap_events(events: pl.LazyFrame, number_events_swap: int) -> pl.LazyFrame:
    number_events = events.select(pl.count()).collect()[0, 0]
    new_events = events.collect()
    for _ in range(number_events_swap):
        row_num = randrange(number_events)
        new_events = (
            new_events.slice(0, row_num - 1)
            .vstack(new_events.slice(row_num + 1, 1))
            .vstack(new_events.slice(row_num, 1))
            .vstack(new_events.slice(row_num + 2, None))
        )

    return new_events.lazy()



"""
data classes
"""

from dataclasses import dataclass, field
from typing import List


@dataclass(eq=True, frozen=True)
class User:
    """
    information about a user required for accessing the API
    """
    user_id: str
    organization_id: str
    api_key: str
    role: str


@dataclass(eq=True, frozen=True)
class SharedInformation:
    """
    sharing information of a resource
    """
    user_ids: List[str] = field(default_factory=list)
    project_ids: List[str] = field(default_factory=list)
    organization_ids: List[str] = field(default_factory=list)


@dataclass(eq=True, frozen=True)
class Resource:
    """
    a generic resource depicting logs, models, shiny dashboards etc.
    """
    name: str
    id: str


class Log(Resource):
    pass


class ShinyDashboard(Resource):
    owner: str
    shared_information: SharedInformation = field(default_factory=SharedInformation)



"""
generic api requests including authorization
"""

from typing import Optional

import requests

from apm4py.decorators import handle_response, expect_json
from apm4py.structures import User


def _create_authorization_header(method: str, token: str) -> dict:
    return {"Authorization": f"{method} {token}"}


def get_token(scheme: str, host: str, instance: int, **kwargs) -> str:
    headers = {**kwargs.pop("headers", dict())}

    port = "" if host != "localhost" else ":4424"

    r = requests.get(
        f"{scheme}://{host}{port}/instance/{instance}/user/analyst/token",
        headers=headers,
        **kwargs,
    )
    r.raise_for_status()
    token = r.content.decode()
    return token


# TODO consider certificate passing for TLS
# TODO consider letting kwargs replace authentication header
class API:
    """An api for a specific user at a Lana deployment.

    All required information to make authenticated requests to the api are
    passed during construction and stored. Named request methods are
    provided as wrappers around requests library methods, that add required
    header fields. Additional headers to the requests library methods can be
    passed as keyword arguments.

    When constructed a request to the lana api is made in order to
    retrieve user information.

    Attributes:
        url (str):
            A string denoting the URL of the lana api, including the
            application root.
        user:
            A User dataclass encapsulating the user of the api information.
        headers:
            A dictionary representing the authorization header used for every
            request by default.
    """

    # TODO document
    def __init__(
        self,
        scheme: str,
        host: str,
        instance: int = 1,
        token: Optional[str] = None,
        port: Optional[int] = None,
        application_root: Optional[str] = None,
        **kwargs,
    ):
        """Construct a configured Appian API.

        Args:
            scheme:
                A string denoting the scheme of the api URL.
            host:
                A string denoting the APM API host.
            token:
                A string denoting the user authentication token without the
                preceding "API-Key".
            port:
                (optional) A string or integer denoting the port of the
                lana api.
            application_root:
                (optional) A string denoting the application root. Only required
                if your lana api is placed outside the URL root, e.g. "/lana-api"
                instead of "/". Has to start with a slash.
            **kwargs:
                Keyword arguments to pass to requests for the initial
                request retrieving user information.
        """

        self.url = (
            f"{scheme}://{host}"
            + (f":{port}" if port else "")
            + (f":{application_root}" if application_root else "").replace("//", "/")
        ).strip("/")
        self.token = token if token else get_token(scheme, host, instance)
        auth_method = "Bearer" if not "lanalabs.com" in self.url else "API-Key"
        self.headers = _create_authorization_header(auth_method, self.token)
        self.user = self.get_user(**kwargs)

    @expect_json
    @handle_response
    def get_user_information(self, **kwargs) -> dict:
        r = requests.get(
            self.url + "/api/users/by-token", headers=self.headers, **kwargs
        )
        r.raise_for_status()
        return r

    def get_user(self, **kwargs) -> User:
        user_info = self.get_user_information(**kwargs)
        return User(
            user_id=user_info.get("id"),
            organization_id=user_info.get("organizationId"),
            api_key=user_info.get("apiKey"),
            role=user_info.get("role"),
        )

    def _request(self, method, route, headers=None, additional_headers=None, **kwargs):
        headers = {
            **self.headers,
            **(additional_headers or dict()),
            **(headers or dict()),
        }
        return requests.request(method, self.url + route, headers=headers, **kwargs)

    @handle_response
    def get(self, route, additional_headers=None, **kwargs):
        return self._request("GET", route, additional_headers, **kwargs)

    @handle_response
    def post(self, route, additional_headers=None, **kwargs):
        return self._request("POST", route, additional_headers, **kwargs)

    @handle_response
    def patch(self, route, additional_headers=None, **kwargs):
        return self._request("PATCH", route, additional_headers, **kwargs)

    @handle_response
    def delete(self, route, additional_headers=None, **kwargs):
        return self._request("DELETE", route, additional_headers, **kwargs)



"""
functions to create and combine trace filters
"""

from dateutil.parser import parse
from collections.abc import Iterable


def combine_filters(*filters) -> list:
    """
    Create a list of combined filters.

    Args:
        *filters:
            Filter arguments to be combined into a trace filter sequence. They
            can be either a single filter as a dictionary or multiple filters in
            a list.

    Returns:
        A list containing the combined filters.
    """
    combined_filter = []

    for trace_filter in filters:
        if isinstance(trace_filter, Iterable):
            combined_filter += trace_filter
        else:
            combined_filter += [trace_filter]

    return combined_filter


def create_timespan_filter(start: str, end: str) -> dict:
    """
    Create a timespan filter to be used in a trace filter sequence.

    Args:
        start:
            A string denoting the start timestamp. The local time is used,
            it is recommended to use ISO-8601 date time formats (e.g.
            'YYYY-MM-DDThh:mm:ss'), but every format that dateutil can parse
            works.
        end:
            A string denoting the end timestamp. The local time is used,
            it is recommended to use ISO-8601 date time formats (e.g.
            'YYYY-MM-DDThh:mm:ss'), but every format that dateutil can parse
            works.

    Returns:
        A dictionary containing the filter.
    """
    return {
        "startInRange": False,
        "endInRange": False,
        "type": "timeRangeFilter",
        "from": parse(start).timestamp() * 1000,
        "to": parse(end).timestamp() * 1000,
    }


def create_attribute_filter(attribute_name: str, values: list) -> dict:
    """
    Create a categorical attribute filter to be used in a trace filter sequence.

    Args:
        attribute_name:
            A string denoting the name of the attribute.
        values:
            A list of values to be filtered.

    Returns:
        A dictionary containing the filter.
    """
    return {
        "type": "attributeFilter",
        "attributeName": attribute_name,
        "values": values,
    }


def create_numeric_attribute_filter(
    attribute_name: str, value_min: float, value_max: float
) -> dict:
    """
    Create a numeric attribute filter to be used in a trace filter sequence.

    Args:
        attribute_name:
            A string denoting the name of the attribute.
        value_min:
            An integer or float denoting the minimum value.
        value_max:
            An integer or float denoting the maximum value.

    Returns:
        A dictionary containing the filter.
    """
    return {
        "type": "numericAttributeFilter",
        "attributeName": attribute_name,
        "min": value_min,
        "max": value_max,
    }


def create_variant_slider_filter(
    min_variant_group: int, max_variant_group: int
) -> dict:
    """
    Create a variant slider filter to be used in a trace filter sequence.

    Args:
        min_variant_group:
            An integer denoting the variant group on the lower bound.
        max_variant_group:
            An integer denoting the variant group on the upper bound.

    Returns:
        A dictionary containing the filter.
    """
    return {
        "type": "variantSliderFilter",
        "min": min_variant_group,
        "max": max_variant_group,
    }


def create_activity_filter(activity: str, include: bool = True) -> dict:
    """
    Create an activity filter to be used in a trace filter sequence.

    Args:
        activity:
            A string denoting the activity to filter.
        include:
            A boolean denoting if the activity should be included or excluded.

    Returns:
        A dictionary containing the filter.
    """
    return {"type": "activityFilter", "activity": activity, "inverted": not include}


def create_activity_filters(include: list, exclude: list = []) -> list:
    """
    Create a list of activity filters to be used in a trace filter sequence.

    Args:
        include:
            A list of strings denoting the activities to include.
        exclude:
            A list of strings denoting the activities to exclude.

    Returns:
        A list containing the activity filters.
    """
    return [create_activity_filter(activity) for activity in include] + [
        create_activity_filter(activity, include=False) for activity in exclude
    ]


def create_follower_filter(
    pre: str, succ: str, direct_follower=False, include=True
) -> dict:
    """
    Create a follower filter to be used in a trace filter sequence.

    Args:
        pre:
            A string denoting the predecessor activity of the follower relation.
        succ:
            A string denoting the successor activity of the follower relation.
        direct_follower:
            A boolean denoting if the activities have to directly
            follow each other.
        include:
            A boolean denoting if the follower relation should be included
            or excluded.

    Returns:
        A list containing the activity filters.
    """
    mapping = {"Start": "__Appian_START__", "End": "__Appian_END__"}

    return {
        "type": "followerFilter",
        "pre": mapping.get(pre, pre),
        "succ": mapping.get(succ, succ),
        "direct": direct_follower,
        "inverted": not include,
    }



"""
decorators for handling api request responses
"""

import functools
import json
from typing import List

import requests


def expect_json(method):
    @functools.wraps(method)
    def parse_json(*args, **kwargs) -> List[dict]:
        resp = method(*args, **kwargs)
        try:
            jsn = resp.json()
        except json.decoder.JSONDecodeError as e:
            mime_type = resp.headers.get('Content-Type')
            raise Exception(f'Expected mime-type application/json got {mime_type}')
        return jsn
    return parse_json


# TODO: introduce proper logging
def handle_response(method):
    @functools.wraps(method)
    def wrapper(*args, **kwargs) -> requests.Response:
        resp = method(*args, **kwargs)
        try:
            resp.raise_for_status()
        except requests.exceptions.HTTPError as e:
            print(e)
        return resp
    return wrapper



"""
All functional parts of the API are combined in the class ProcessMiningAPI.

"""

from apm4py.aggregation import AggregationAPI
from apm4py.logs import LogsAPI
from apm4py.resources import ResourceAPI


class ProcessMiningAPI(LogsAPI, AggregationAPI, ResourceAPI):
    """The Python API for Appian Process Mining.

    It allows programmatic management of all lana resources.

    All required information to make authenticated requests to the api are
    passed during construction and stored. Named request methods are
    provided as wrappers around requests library methods, that add required
    header fields. Additional headers to the requests library methods can be
    passed as keyword arguments.

    Attributes:
        url (str):
            The base url of the api (scheme, host and port).
        user (User):
            A User dataclass encapsulating the user of the api information.
        headers (dict):
            The authorization header used for every request by default.
    """

    pass



import dataclasses
import json
import math
import os
from pathlib import Path
import polars as pl
import shutil
import sys
import typer
from typing import List, Optional
from typing_extensions import Annotated

import apm4py
from apm4py.event_log import EventLog
import apm4py.log_grower
from apm4py.utils import get_file_by_pattern
from apm4py.log_io import load_log
from apm4py.log_manipulation import *

app = typer.Typer()


@app.command()
def upload(
    data_folder: str = typer.Argument(help="Folder where data sets are stored"),
    name: Optional[str] = typer.Option(None, help="Name of the event log"),
    profile: str = typer.Option("default", help="Profile of the Process Mining host"),
    chunk_size: int = typer.Option(
        250000, help="Number of lines to upload in one chunk"
    ),
):
    profile_config = get_profile_config(profile)
    api = apm4py.create_api(**profile_config)
    event_path = get_file_by_pattern(data_folder, "*[eE]vent*.csv")
    event_semantics = get_file_by_pattern(data_folder, "*[eE]vent*.json")
    case_file = get_file_by_pattern(data_folder, "*[cC]ase*.csv")
    case_semantics = get_file_by_pattern(data_folder, "*[cC]ase*.json")

    name = name if name is not None else Path(event_path).stem
    api.upload_event_log_file(
        name=name,
        event_file_path=event_path,
        event_semantics_path=event_semantics,
        case_file_path=case_file,
        case_semantics_path=case_semantics,
        chunk_size=chunk_size,
        show_progress=True,
    )


@app.command()
def list_logs(
    profile: str = typer.Option("default", help="Profile of the Process Mining host"),
):
    profile_config = get_profile_config(profile)
    api = apm4py.create_api(**profile_config)
    logs = api.list_logs()

    if len(logs) > 0:
        pl.Config.set_fmt_str_lengths(100)
        print(
            pl.from_dicts(logs)
            .with_columns(pl.from_epoch("insertedAt", time_unit="ms"))
            .select(["id", "name", "insertedAt"])
            .sort("insertedAt", descending=True)
        )

    else:
        print("No logs available")


def get_profile_config(profile: str):
    apm_dir = os.path.join(os.environ.get("HOME"), ".apm")
    apm_profile_path = os.path.join(os.environ.get("HOME"), ".apm", profile)
    if not os.path.exists(apm_profile_path):
        if not os.path.exists(apm_dir):
            os.mkdir(apm_dir)

        print(f"Profile {profile} does not exists. Please create a profile first.")
        profile = typer.prompt("Name of your profile", default="default")
        if profile == "":
            profile = "default"

        host = typer.prompt("Hostname")
        scheme = typer.prompt("Scheme", default="https")
        port = typer.prompt("Port", default="auto")
        token = typer.prompt("API Token", default="None")
        instance = typer.prompt("Instance", default=2)

        profile_config = {"host": host, "scheme": scheme, "instance": instance}

        if port != "auto":
            profile_config["port"] = port

        if token != "None":
            profile_config["token"] = token

        apm_profile_path = os.path.join(os.environ.get("HOME"), ".apm", profile)
        with open(apm_profile_path, "w") as profile_file:
            json.dump(profile_config, profile_file)

    else:
        with open(apm_profile_path, "r") as profile_file:
            profile_config = json.loads(profile_file.read())

    return profile_config


@app.command()
def grow_log(
    orig_data_folder: Annotated[
        Path, typer.Argument(help="Folder for the original dataset")
    ],
    dataset_collection: Annotated[
        str, typer.Option(help="Name of the dataset collection (P2P_Benchmark|)")
    ] = None,
    output_csv: Annotated[
        bool, typer.Option(help="Output CSV files instead of parquet")
    ] = False,
    output_feather: Annotated[
        bool, typer.Option(help="Output feather files instead of parquet")
    ] = False,
    feather_compression: Annotated[
        str, typer.Option(help="uncompressed|lz4|zstd")
    ] = "uncompressed",
    compression_level: Annotated[
        int, typer.Option(help="File compression level")
    ] = None,
    data_folder: Annotated[
        str, typer.Option(help="Output folder for datasets")
    ] = "./data",
    dataset: Annotated[
        Optional[List[str]],
        typer.Option(help="Only build a specific dataset from collection"),
    ] = None,
    partitions: Annotated[
        int, typer.Option(help="Overwrite num partitions with given number")
    ] = None,
    use_internal_format: Annotated[
        bool,
        typer.Option(help="Use the internally used format with numerical Case IDs"),
    ] = False,
    csv_separator: Annotated[str, typer.Option()] = ",",
    growth_factor: Annotated[
        int,
        typer.Option(
            help="Frequency of appending the log to itself. Only used when no dataset collection is selected.",
        ),
    ] = None,
):
    orig_log = load_log(
        data_folder=orig_data_folder,
        csv_separator=csv_separator,
        use_internal_format=use_internal_format,
    )
    new_dataset_path = apm4py.log_grower.grow_log(
        events_original=orig_log.events.collect(),
        events_case_id=orig_log.caseid_events(),
        cases_original=orig_log.cases.collect(),
        cases_case_id=orig_log.caseid_cases(),
        variants_original=orig_log.variants,
        dataset_collection=dataset_collection,
        output_csv=output_csv,
        output_feather=output_feather,
        feather_compression=feather_compression,
        compression_level=compression_level,
        data_folder=data_folder,
        dataset=dataset,
        partitions=partitions,
        use_internal_format=use_internal_format,
        growth_factor=growth_factor,
    )

    orig_log.save_semantics(new_dataset_path)

    # shutil.copy(case_semantics_path, new_dataset_path)
    # shutil.copy(event_semantics_path, new_dataset_path)


@app.command()
def manipulate_log(
    orig_data_folder: Annotated[
        Path, typer.Argument(help="Folder for the original dataset")
    ],
    data_folder: Annotated[
        str, typer.Option(help="Output folder for datasets")
    ] = "./data",
    dataset_name: Annotated[
        str,
        typer.Option(help="Name of the newly generated dataset"),
    ] = "manipulated_dataset",
    csv_separator: Annotated[str, typer.Option()] = ",",
    use_internal_format: Annotated[
        bool,
        typer.Option(help="Use the internally used format with numerical Case IDs"),
    ] = False,
    number_variants: Annotated[
        int, typer.Option(help="Number of variants yielded by the manipulation")
    ] = None,
):
    orig_log = load_log(
        data_folder=orig_data_folder,
        csv_separator=csv_separator,
        use_internal_format=use_internal_format,
    )

    orig_num_variants = orig_log.compute_variants().collect().height

    if number_variants is not None:
        variant_diff = number_variants - orig_num_variants
        if variant_diff <= 1:
            print(
                f"""
                The event log already has {orig_num_variants} variants.
                You wanted to reach {number_variants}, so there's nothing to do.
                Use --number-variants to yield a larger number of variants.
                """
            )

        new_log = dataclasses.replace(orig_log)
        new_log_path = os.path.join(data_folder, dataset_name)

        while variant_diff > 0:
            num_evts_per_step = math.ceil(variant_diff / 3)

            new_events = new_log.events
            new_events = remove_events(new_events, num_evts_per_step)
            new_events = swap_events(new_events, num_evts_per_step)
            new_events = duplicate_events(new_events, num_evts_per_step)
            new_log.events = new_events

            new_num_variants = new_log.compute_variants().collect().height

            variant_diff = number_variants - new_num_variants

    if new_num_variants is None:
        new_num_variants = new_log.compute_variants().collect().height

    print(
        f"""
          Saving new event log with 
          {new_num_variants} variants (orig: {orig_num_variants}) and
          {new_log.events.collect().height} events (orig: {orig_log.events.collect().height})
          to {new_log_path}
          """
    )

    new_log.save_log(new_log_path)


def main():
    sys.exit(app())


if __name__ == "__main__":
    main()



"""
resource management api requests
"""

import re
from typing import List, Union

from requests import Response

from apm4py.api import API
from apm4py.decorators import expect_json


class ResourceAPI(API):
    @expect_json
    def list_resources(self, kind: str, **kwargs) -> list:
        """
        Lists all resources of a kind.

        Args:
            kind:
                A string denoting the resource type.
            **kwargs:
                Keyword arguments passed to requests functions.
        """
        return self.get(f"/api/{kind}", **kwargs)

    def get_resource_ids(self, kind: str, contains: str, **kwargs) -> List[str]:
        """
        Get ids of resource with names are matched by regular expression

        Args:
            kind:
                A string denoting the resource type.
            contains:
                A string denoting a regular expression matched against the resource
                names.
            **kwargs:
                Keyword arguments passed to requests functions.

        Returns:
            a list of strings representing resource ids
        """
        resources = self.list_resources(kind, **kwargs)
        rc = re.compile(contains)

        return [
            resource.get("id") or resource.get("pageId")
            for resource in resources
            if rc.search(resource.get("name") or resource.get("title"))
        ]

    def get_resource_id(self, kind: str, contains: str, **kwargs) -> str:
        """
        Get id of a resource by its name

        name needs to be unique or an exception is raised
        """
        resource_ids = self.get_resource_ids(kind, contains, **kwargs)

        try:
            [resource_id] = resource_ids
        except ValueError as e:
            raise Exception(
                f"Found {len(resource_ids)} resources with the pattern {contains}"
            )

        return resource_id

    @expect_json
    def describe_resource(
        self, kind: str, contains: str = None, resource_id: str = None, **kwargs
    ) -> dict:
        """
        get description of resource

        Args:
            kind: resource type name
            resource_id: The id of the log, takes precedence over contains
            contains: a regex matching the resource's name, matching several names raises an exception

        Returns:

        """
        resource_id = resource_id or self.get_resource_id(kind, contains, **kwargs)
        return self.get(f"/api/{kind}/{resource_id}")

    @expect_json
    def create_resource(self, kind: str, json: Union[list, dict, str], **kwargs):
        return self.post(f"/api/{kind}", json=json, **kwargs)

    def delete_resource(self, kind, id_: str, **kwargs) -> Response:
        """
        delete a log by its id
        """
        return self.delete(f"/api/{kind}/{id_}", **kwargs)

    def delete_resources(
        self, kind: str, contains: str = None, ids: List[str] = None, **kwargs
    ) -> List[Response]:
        """
        deletes one or multiple logs matching the passed regular expression
        """
        ids = ids or self.get_resource_ids(kind, contains, **kwargs)
        return [self.delete_resource(kind, id_) for id_ in ids]

    def connect_resources(self, dct, **kwargs) -> Response:
        return self.post("/api/v2/resource-connections", json=dct, **kwargs)

    def connect_model(self, log_id, model_id, **kwargs):
        dct = {"log_id": log_id, "model_id": model_id}
        return self.connect_resources(dct, **kwargs)

    def connect_working_schedule(self, log_id, working_schedule_id, **kwargs):
        dct = {"log_id": log_id, "working_schedule_id": working_schedule_id}
        return self.connect_resources(dct, **kwargs)



"""
functions to create the aggregation request metric
"""

from typing import Optional, Any

from apm4py.aggregation.utils import _raise


def create_frequency(kind: str) -> dict:
    return {"type": kind} if kind == "frequency" else dict()


def create_duration(kind: str) -> dict:
    return {"type": kind} if kind == "duration" else dict()


def create_attribute(att: str) -> dict:
    return (
        {"type": "attribute", "attribute": att}
        if att not in ["frequency", "duration"]
        else dict()
    )


def _is_summable(metric: dict, arg: Any) -> bool:
    return metric and arg and metric.get("type") != "frequency"


def add_percentile(metric: dict, p: float) -> dict:
    return (
        {**metric, **{"aggregationFunction": {"type": "percentile", "percentile": p}}}
        if _is_summable(metric, p)
        else dict()
    )


def add_aggregation(metric: dict, agg: str) -> dict:
    return (
        {**metric, **{"aggregationFunction": agg}}
        if _is_summable(metric, agg)
        else dict()
    )


def create_summable(kind):
    return create_duration(kind) or create_attribute(kind)


def aggregate_summable(summable, aggregator, percentile):
    if aggregator and percentile:
        raise Exception("Either aggregator or percentile can be not None, " "not both.")
    return add_aggregation(summable, aggregator) or add_percentile(summable, percentile)


def create_metric(
    kind: str, aggregator: Optional[str] = None, percentile: Optional[float] = None
) -> dict:
    """
    Create a dictionary containing the metric in a format necessary
    for the aggregation API request.

    Args:
        kind:
            A string denoting the metric to use. For the value
            "frequency", a frequency metric is returned and for "duration" a
            duration metric is returned. Otherwise the value is interpreted
            as a numeric attribute metric. For these metrics and duration
            metrics an aggregator or percentile needs to be passed.
        aggregator:
            An optional string denoting the aggregation function to use.
            Can be one of ["min", "max", "sum", "mean", "median", "variance",
            "standardDeviation"].
        percentile:
            An optional float denoting the percentile to use if instead of
            the available aggregation types listed for aggregation_function
            a percentile aggregation should be used.

    Returns:
        A dictionary containing the metric in the right format for the request.
    """
    return (
        aggregate_summable(create_summable(kind), aggregator, percentile)
        or create_frequency(kind)
        or _raise(
            Exception(
                f"Impossible to create metric for {kind} without "
                f"aggregator or percentile."
            )
        )
    )
